/*
 * Copyright (c) 2002 Ian Dowse.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>
#include <sys/disklabel.h>
#include <sys/time.h>

#include <ufs/ufs/dinode.h>
#include <ufs/ufs/dir.h>
#include <ufs/ffs/fs.h>

#include <assert.h>
#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "swapfs_util.h"

#define DPRINTF(x) do { \
	if (debug) { \
		printf x; \
		fflush(stdout); \
	} \
} while (0)

static int debug;
static int nflag;
static int yflag;

static char *dev;
static int dev_fd;
static int fsorder;

static struct {
	int ncg;
	int ninodes;
	int nindir;
	int ndirs;
	int ndirblks;
} stats;

static char sbbuf[SBSIZE];
static struct fs *fs;

void do_cg(int cg);
void do_csum(void);
void do_inode(int inum , struct dinode *dip);
void do_indblk(struct dinode *di, int level, int fsb, int lbn);
void do_dirblk(struct dinode *di, int fsb, int lbn);

void devread(void *buf, size_t nbytes, off_t off);
void devwrite(void *buf, size_t nbytes, off_t off);
static void usage(void);

int
main(int argc, char **argv)
{
	char buf[SBSIZE];
	int ch, i;

	while ((ch = getopt(argc, argv, "dny")) != -1) {
		switch (ch) {
		case 'd':
			debug++;
			break;
		case 'n':
			nflag++;
			break;
		case 'y':
			yflag++;
			break;
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;

	if (argc != 1)
		usage();

	dev = *argv;
	if ((dev_fd = open(dev, nflag ? O_RDONLY : O_RDWR)) < 0)
		err(1, "%s: open", dev);
	DPRINTF(("Reading superblock\n"));
	devread(sbbuf, SBSIZE, SBOFF);
	fs = (struct fs *)sbbuf;

	if (fs->fs_magic == FS_MAGIC)
		fsorder = FSORDER_HOST;
	else if (fs->fs_magic == bswap32(FS_MAGIC))
		fsorder = FSORDER_SWAP;
	else
		errx(1, "%s: bad magic (0x%08x) or unknown byte order", dev,
		    fs->fs_magic);

	printf("Working with %s\n", dev);
	printf("Converting from %s (%s) to %s (%s)\n",
	    (fsorder == FSORDER_HOST) ? "HOST" : "SWAPPED",
	    FSORDER(fsorder), (fsorder == FSORDER_HOST) ? "SWAPPED" : "HOST",
	    FSORDER((fsorder == FSORDER_HOST) ? FSORDER_SWAP : FSORDER_HOST));

	/* Set up a host-order superblock. */
	if (fsorder != FSORDER_HOST)
		swapfs_sb(NULL, sbbuf, SBSIZE);
	assert(fs->fs_magic == FS_MAGIC);

	/* Sanity checks. */
	for (i = 0; i < FSMAXSNAP; i++)
		if (fs->fs_snapinum[i] != 0)
			errx(1, "Sorry, snapshots are not supported yet.");
	if (fs->fs_inodefmt < FS_44INODEFMT)
		errx(1, "Sorry, old inode format is not supported.");
	if (fs->fs_postblformat != FS_DYNAMICPOSTBLFMT)
		errx(1, "Sorry, cylinder group format is not supported.");

	if (!fs->fs_clean)
		warnx("***WARNING*** filesystem is not marked clean.");

	if (!yflag) {
		printf("\n");
		printf("Warning: do not interrupt while running. If the\n");
		printf("conversion fails or is aborted, the filesystem\n");
		printf("will be destroyed!\n");
		printf("Begin conversion [y/n]? ");
		fflush(stdout);

		i = fgetc(stdin);
		if (i != 'Y' && i != 'y')
			errx(0, "Filesystem left unchanged.");
	}

	/* Mark as unclean while working */
	DPRINTF(("Marking filesystem as dirty\n"));
	devread(buf, SBSIZE, SBOFF);
	((struct fs *)buf)->fs_clean = 0;
	devwrite(buf, SBSIZE, SBOFF);

	/* Swap the superblock. */
	devread(buf, SBSIZE, SBOFF);
	swapfs_sb(fs, buf, SBSIZE);

	/* Summary information. */
	do_csum();

	printf("cg:\n");
	for (i = 0; i < fs->fs_ncg; i++) {
		printf(" %d", i);
		fflush(stdout);
		do_cg(i);
	}
	printf("\n");

	/* Reset the clean flag, and write the swapped superblock */
	if (fs->fs_clean) {
		DPRINTF(("Marking filesystem as clean\n"));
	}
	((struct fs *)buf)->fs_clean = fs->fs_clean;
	devwrite(buf, SBSIZE, SBOFF);

	printf("Conversion stats:\n");
	printf("\tcylinder groups:    %d\n", stats.ncg);
	printf("\tinodes:             %d\n", stats.ninodes);
	printf("\tindirection blocks: %d\n", stats.nindir);
	printf("\tdirectories:        %d\n", stats.ndirs);
	printf("\tdirectory blocks:   %d\n", stats.ndirblks);
	printf("\nFilesystem successfully converted.\n");
	return (0);
}

void
do_csum(void)
{
	static void *buf;
	static int buflen;
	struct csum *csp;
	int datalen;

	datalen = fs->fs_cssize;
	if (buflen < datalen) {
		if ((buf = realloc(buf, datalen)) == NULL)
			err(1, "do_cs: realloc %d -> %d", buflen, datalen);
		buflen = datalen;
	}

	devread(buf, datalen, (off_t)fs->fs_csaddr * (off_t)fs->fs_fsize);
	csp = (struct csum *)buf;
	for (; (char *)csp < (char *)buf + buflen; csp++)
		swapfs_csum(fs, csp);
	devwrite(buf, datalen, (off_t)fs->fs_csaddr * (off_t)fs->fs_fsize);
}

void
do_cg(int cgnum)
{
	static void *buf;
	static int buflen;
	struct cg *cgp, cg;
	struct dinode *dip;
	int dlower, dupper, datalen, fsb, i;

	stats.ncg++;
	DPRINTF(("cg %d\n", cgnum));

	dlower = cgsblock(fs, cgnum);
	dupper = cgdmin(fs, cgnum);
	datalen = (dupper - dlower) * fs->fs_fsize;

	if (buflen < datalen) {
		if ((buf = realloc(buf, datalen)) == NULL)
			err(1, "do_cg: realloc %d -> %d", buflen, datalen);
		buflen = datalen;
	}

	/* Read in all the cylinder group meta-data in one chunk. */
	devread(buf, datalen, (off_t)dlower * (off_t)fs->fs_fsize);

	/* The backup superblock is right at the start; swap it. */
	swapfs_sb(fs, buf, SBSIZE);

	/* Now find the cylinder group structure and swap it. */
	fsb = cgtod(fs, cgnum);
	assert(fsb >= dlower && fsb < dupper);
	cgp = (struct cg *)((char *)buf + (fsb - dlower) * fs->fs_fsize);

	cg = *cgp;
	if (fsorder != FSORDER_HOST)
		swapfs_cg(fs, NULL, &cg, sizeof(cg));
	assert(cg.cg_magic == CG_MAGIC);
	assert(cg.cg_cgx == cgnum);

	swapfs_cg(fs, &cg, (void *)cgp, fs->fs_cgsize);

	/* Next do the inodes. */
	for (i = cgnum * fs->fs_ipg; i < (cgnum + 1) * fs->fs_ipg; i++) {
		fsb = ino_to_fsba(fs, i);
		assert(fsb >= dlower && fsb < dupper);
		dip = &((struct dinode *)((char *)buf +
		    (fsb - dlower) * fs->fs_fsize))[ino_to_fsbo(fs, i)];
		assert((char *)dip >= (char *)buf &&
		    (char *)dip + sizeof(*dip) <= (char *)buf + buflen);

		do_inode(i, dip);
	}

	/* Write out all the cylinder group meta-data. */
	devwrite(buf, datalen, (off_t)dlower * (off_t)fs->fs_fsize);

}

void
do_inode(int inum, struct dinode *dip)
{
	struct dinode di;
	int i, lbn, lbcount;

	di = *dip;
	if (fsorder != FSORDER_HOST)
		swapfs_dinode(fs, &di);
	assert(di.di_size <= fs->fs_maxfilesize);

	/* Check for free inodes. */
	if (di.di_mode == 0)
		return;

	stats.ninodes++;
	swapfs_dinode(fs, dip);

	DPRINTF(("inode %d: mode %07o size %lld\n", inum, di.di_mode,
	    (long long)di.di_size));
	switch (di.di_mode & IFMT) {
	case IFDIR:
		stats.ndirs++;
		/* Swap direct blocks */
		for (lbn = 0; lbn < NDADDR; lbn++)
			if (di.di_db[lbn] != 0)
				do_dirblk(&di, di.di_db[lbn], lbn);

		/* Go on to indirect blocks. */
		break;
	case IFREG:
		/* Go on to indirect blocks. */
		break;
	default:
		/* Nothing more to do. */
		return;
	}

	lbn = NDADDR;
	lbcount = 1;
	for (i = 0; i < NIADDR; i++) {
		lbcount *= NINDIR(fs);
		if (di.di_ib[i] != 0)
			do_indblk(&di, i, di.di_ib[i], lbn);
		lbn += lbcount;
	}
}

void
do_indblk(struct dinode *di, int level, int fsb, int lbn)
{
	static char *lbuf[NIADDR];
	static int buflen[NIADDR];
	void *buf;
	ufs_daddr_t *blkp;
	int datalen, i, lbcount;

	assert(fsb % fs->fs_frag == 0);
	assert(fsb < fs->fs_size);

	stats.nindir++;
	DPRINTF(("indir: fsb %d lbn %d level %d\n", fsb, lbn, level));

	datalen = fs->fs_bsize;
	if (buflen[level] < datalen) {
		if ((lbuf[level] = realloc(lbuf[level], datalen)) == NULL)
			err(1, "do_cg: realloc %d -> %d", buflen[level],
			    datalen);
		buflen[level] = datalen;
	}
	buf = lbuf[level];

	/* Swap this indirect block and write it out. */
	devread(buf, datalen, (off_t)fsb * (off_t)fs->fs_fsize);
	swapfs_indir(fs, buf, datalen);
	devwrite(buf, datalen, (off_t)fsb * (off_t)fs->fs_fsize);

	/* If necessary, swap back before looking at the block numbers. */
	if (fsorder == FSORDER_HOST)
		swapfs_indir(fs, buf, datalen);
	blkp = buf;

	if (level == 0) {
		/* No more levels of indirect blocks. Process directories. */
		if ((di->di_mode & IFMT) != IFDIR)
			return;
		for (i = 0; i < NINDIR(fs); i++)
			if (blkp[i] != 0)
				do_dirblk(di, blkp[i], lbn + i);
		return;
	}

	/* Determine how much the LBA is advanced for each index. */
	lbcount = 1;
	for (i = 0; i < level; i++)
		lbcount *= NINDIR(fs);

	/* Recurse into subblocks. */
	for (i = 0; i < NINDIR(fs); i++)
		if (blkp[i] != 0)
			do_indblk(di, level - 1, blkp[i], lbn + lbcount * i);
}

void
do_dirblk(struct dinode *di, int fsb, int lbn)
{
	static void *buf;
	static int buflen;
	char hobuf[DIRBLKSIZ];
	int datalen, dirend, i;

	assert(fsb < fs->fs_size);
	DPRINTF(("dir: fsb %d lbn %d\n", fsb, lbn));

	datalen = dblksize(fs, di, lbn);
	if (buflen < datalen) {
		if ((buf = realloc(buf, datalen)) == NULL)
			err(1, "do_dirblk: realloc %d -> %d", buflen, datalen);
		buflen = datalen;
	}

	dirend = datalen;
	if (di->di_size < (off_t)lbn * (off_t)fs->fs_bsize + datalen)
		dirend = di->di_size - (off_t)lbn * (off_t)fs->fs_bsize;

	devread(buf, datalen, (off_t)fsb * (off_t)fs->fs_fsize);
	for (i = 0; i < dirend; i += DIRBLKSIZ) {
		bcopy((char *)buf + i, hobuf, DIRBLKSIZ);
		if (fsorder != FSORDER_HOST)
			swapfs_dir(fs, NULL, hobuf, DIRBLKSIZ);
			
		swapfs_dir(fs, hobuf, (char *)buf + i, DIRBLKSIZ);
		stats.ndirblks++;
	}
	devwrite(buf, datalen, (off_t)fsb * (off_t)fs->fs_fsize);
}

static void
usage()
{

	(void)fprintf(stderr,
            "usage: swapfs [-dny] special\n");
	exit(1);
}

void
devread(void *buf, size_t nbytes, off_t off)
{
	ssize_t nread;

	nread = pread(dev_fd, buf, nbytes, off);
	if (nread < 0)
		err(1, "%s: read %d@%lld", dev, (int)nbytes, (long long)off);
	if (nread != nbytes)
		errx(1, "%s: read %d@%lld short read (%d)", dev, (int)nbytes,
		    (long long)off, (int)nread);
}

void
devwrite(void *buf, size_t nbytes, off_t off)
{
	ssize_t nwritten;

	if (nflag)
		return;

	nwritten = pwrite(dev_fd, buf, nbytes, off);
	if (nwritten < 0)
		err(1, "%s: write %d@%lld", dev, (int)nbytes, (long long)off);
	if (nwritten != nbytes)
		errx(1, "%s: write %d@%lld short write (%d)", dev, (int)nbytes,
		    (long long)off, (int)nwritten);
}
