/*
 * Copyright (c) 2002 Ian Dowse.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>

#include <ufs/ufs/dinode.h>
#include <ufs/ufs/dir.h>
#include <ufs/ffs/fs.h>

#include <assert.h>
#include <err.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "swapfs_util.h"

struct structfield {
	int offset;
	int bytes;
	int count;
};
#define SFIELD(sname, name, count) {offsetof(struct sname, name), \
    sizeof(((struct sname *)0)->name), count}
#define SFEND {-1}

static struct structfield sb_fields[] =  {
	SFIELD(fs, fs_firstfield, 48),
	/* fs_cstotal done specially. */
	SFIELD(fs, fs_fmod, 4),
	SFIELD(fs, fs_cgrotor, 1),
	SFIELD(fs, fs_cpc, 1),
	/* fs_opostbl done specially. */
	SFIELD(fs, fs_snapinum[0], FSMAXSNAP),
	SFIELD(fs, fs_avgfilesize, 2),
	/* Ignore fs_sparecon. */
	SFIELD(fs, fs_pendingblocks, 5),
	SFIELD(fs, fs_maxfilesize, 3),
	SFIELD(fs, fs_state, 6),
	SFEND
};

static struct structfield csum_fields[] =  {
	SFIELD(csum, cs_ndir, 1),
	SFIELD(csum, cs_nbfree, 1),
	SFIELD(csum, cs_nifree, 1),
	SFIELD(csum, cs_nffree, 1),
	SFEND
};

static struct structfield dinode_fields[] =  {
	SFIELD(dinode, di_mode, 2),
	/* ignore di_u */
	SFIELD(dinode, di_size, 1),
	SFIELD(dinode, di_atime, 6 + NDADDR + NIADDR + 5),
	SFEND
};

static struct structfield cg_fields[] =  {
	SFIELD(cg, cg_firstfield, 4),
	SFIELD(cg, cg_ncyl, 2),
	SFIELD(cg, cg_ndblk, 1),
	/* cg_cs done specially. */
	SFIELD(cg, cg_rotor, 3 + MAXFRAG + 8),
	/* cg_sparecon ignored. */
	SFEND
};

static struct structfield dir_fields[] =  {
	SFIELD(direct, d_ino, 1),
	SFIELD(direct, d_reclen, 1),
	/* d_type, d_namlen, d_name need no swapping. */
	SFEND
};

static void swapfs_struct(struct structfield *sf, void *buf, int buflen);

/*
 * Swap a superblock in `buf'. `fs' should point to a host-order
 * superblock which is used for parameters. If `fs' is NULL, only
 * the main struct fs fields are swapped (i.e. not the extra tables).
 */
void
swapfs_sb(struct fs *fs, void *buf, int buflen)
{
	struct structfield sf[2];
	struct fs *fsp;

	assert(buflen >= SBSIZE);
	assert(fs == NULL || fs->fs_magic == FS_MAGIC);

	/* Swap the main superblock fields. */
	fsp = buf;
	swapfs_struct(sb_fields, fsp, sizeof(*fsp));
	swapfs_struct(csum_fields, &fsp->fs_cstotal, sizeof(fsp->fs_cstotal));

	/* No host-order struct fs. */
	if (fs == NULL)
		return;

	/* Swap the rotational position table. */
	sf[0].offset = (char *)fs_postbl(fs, 0) - (char *)fs;
	sf[0].bytes = 2;
	sf[0].count = fs->fs_cpc * fs->fs_nrpos;
	sf[1].offset = -1;
	swapfs_struct(sf, buf, buflen);
}

/*
 * Swap the `struct csum' pointed to by csp.
 */
void
swapfs_csum(struct fs *fs, struct csum *csum)
{
	swapfs_struct(csum_fields, csum, sizeof(*csum));
}

/*
 * Swap the `struct cg' contained in `buf'. `cg' points to a host-order
 * version of the struct cg, or NULL if none is available.
 */
void
swapfs_cg(struct fs *fs, struct cg *cg, void *buf, int buflen)
{
	struct structfield sf[4];
	struct cg *cgp;

	cgp = buf;

	/* Swap the main structure. */
	swapfs_struct(cg_fields, cgp, sizeof(*cgp));
	swapfs_csum(fs, &cgp->cg_cs);

	/* Cannot swap the rest without a host-order cg. */
	if (cg == NULL)
		return;

	sf[0].offset = cg->cg_btotoff;
	sf[0].bytes = 4;
	sf[0].count = fs->fs_cpg;

	sf[1].offset = cg->cg_boff;
	sf[1].bytes = 2;
	sf[1].count = fs->fs_cpg * fs->fs_nrpos;

	/*
	 * Since the zeroth entry in the cg_clustersumoff array is unused,
	 * cg_clustersumoff is 4 bytes before the real start of the array.
	 */
	sf[2].offset = cg->cg_clustersumoff + 4;
	sf[2].bytes = 4;
	sf[2].count = fs->fs_contigsumsize;

	sf[3].offset = -1;
	swapfs_struct(sf, buf, buflen);
}

/*
 * Swap the `struct dinode' pointed to by dip.
 */
void
swapfs_dinode(struct fs *fs, struct dinode *dip)
{
	swapfs_struct(dinode_fields, dip, sizeof(*dip));
}

/*
 * Swap an indirection block
 */
void
swapfs_indir(struct fs *fs, void *buf, int buflen)
{
	struct structfield sf[2];

	sf[0].offset = 0;
	sf[0].bytes = sizeof(ufs_daddr_t);
	sf[0].count = NINDIR(fs);

	sf[1].offset = -1;
	swapfs_struct(sf, buf, buflen);
}

/*
 * Swap a directory DIRBLKSIZ-size block. `hobuf' should contain a
 * host-order version of the same block. If `hobuf' is NULL, `buf'
 * MUST be in non-host order.
 */
void
swapfs_dir(struct fs *fs, void *hobuf, void *buf, int buflen)
{
	struct direct *dp, *hp;
	int i;

	assert(buflen == DIRBLKSIZ);
	for (i = 0; i < DIRBLKSIZ;) {
		dp = (struct direct *)((char *)buf + i);
		swapfs_struct(dir_fields, dp, DIRBLKSIZ - i);
	
		if (hobuf != NULL)
			hp = (struct direct *)((char *)hobuf + i);
		else
			hp = dp;

		assert(hp->d_reclen != 0 && i + hp->d_reclen <= DIRBLKSIZ);
		i += hp->d_reclen;
	}
}

/*
 * Byte-swap a structure according to the specification in `sf'.
 */
static void
swapfs_struct(struct structfield *sf, void *buf, int buflen)
{
	struct structfield *sfp;
	void *p, *prevp;
	int i;

	prevp = buf;
	for (sfp = sf; sfp->offset != -1; sfp++) {
		for (i = 0; i < sfp->count; i++) {
			p = (char *)buf + sfp->offset + i * sfp->bytes;
			assert(p >= prevp);
			assert((char *)p + sfp->bytes <= (char *)buf + buflen);

			switch (sfp->bytes) {
			case 1:
				break;
			case 2:
				*(u_int16_t *)p = bswap16(*(u_int16_t *)p);
				break;
			case 4:
				*(u_int32_t *)p = bswap32(*(u_int32_t *)p);
				break;
			case 8:
				*(u_int64_t *)p = bswap64(*(u_int64_t *)p);
				break;
			default:
				printf("cannot swap %d-byte field\n",
				     sfp->bytes);
				abort();
			}

			prevp = (char *)p + sfp->bytes;
		}
	}
}
				

u_int16_t
bswap16(u_int16_t x)
{
	x = ((x << 8) & 0xff00) | ((x >> 8) & 0x00ff);
	return (x);
}

u_int32_t
bswap32(u_int32_t x)
{
	x = ((x << 16) & 0xffff0000UL) | ((x >> 16) & 0x0000ffffUL);
	x = ((x << 8) & 0xff00ff00UL) | ((x >> 8) & 0x00ff00ffUL);
	return (x);
}

u_int64_t
bswap64(u_int64_t x)
{
	x = ((x << 32) & 0xffffffff00000000ULL) |
	    ((x >> 32) & 0x00000000ffffffffULL);
	x = ((x << 16) & 0xffff0000ffff0000ULL) |
	    ((x >> 16) & 0x0000ffff0000ffffULL);
	x = ((x << 8) & 0xff00ff00ff00ff00ULL) |
	    ((x >> 8) & 0x00ff00ff00ff00ffULL);
	return (x);
}

